#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mutex.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/hrtimer.h>
#include <linux/ktime.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <linux/videodev2.h>
#include <media/videobuf2-vmalloc.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-device.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-event.h>

struct buffer {
	struct vb2_v4l2_buffer vb;
	struct buffer *next;
};

static int init(void);
static void shutdown(void);

static int ioctl_querycap(struct file *file, void *fh, struct v4l2_capability *cap);
static int ioctl_enum_fmt(struct file *file, void *fh, struct v4l2_fmtdesc *fdesc);
static int ioctl_enum_framesz(struct file *file, void *fh, struct v4l2_frmsizeenum *fsz);
static int ioctl_get_fmt(struct file *file, void *fh, struct v4l2_format *fmt);
static int ioctl_set_fmt(struct file *file, void *fh, struct v4l2_format *fmt);
static int ioctl_try_fmt(struct file *file, void *fh, struct v4l2_format *fmt);
static int ioctl_get_parm(struct file *file, void *fh, struct v4l2_streamparm *fmt);
static int ioctl_set_parm(struct file *file, void *fh, struct v4l2_streamparm *fmt);
static int ioctl_queryctl(struct file *file, void *fh, struct v4l2_queryctrl *ctl);

/* TODO find out exact version where this switched over */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,7,0)
#define ALLOC_CTX_TYPE	void
#else
#define ALLOC_CTX_TYPE	struct device
#endif

static int queue_setup(struct vb2_queue *vbq, unsigned int *nbuf,
		unsigned int *nplanes, unsigned int *sizes, ALLOC_CTX_TYPE **alloc_ctx);
static int buf_prepare(struct vb2_buffer *vb);
static void buf_queue(struct vb2_buffer *vb);
static int start_streaming(struct vb2_queue *vbq, unsigned int count);
static void stop_streaming(struct vb2_queue *vbq);


static int update_frame(int xsz, int ysz);
static int gen_frame(void);
enum hrtimer_restart timer_func(struct hrtimer *timer);


module_init(init);
module_exit(shutdown);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("John Tsiombikas");
MODULE_DESCRIPTION("Dummy V4L2 device");

static struct mutex mutex;
static struct video_device *vdev;

static struct v4l2_device v4l2_dev;
static struct v4l2_file_operations fops;
static struct v4l2_ioctl_ops iops;
static struct vb2_queue vbq;
static struct vb2_ops vbops;
static int width, height;
static unsigned char *frame;
static int frame_size;

static int streaming;
static spinlock_t blist_lock;
static struct buffer *buflist, *buflist_tail;
static int seqno;

static int framerate = 1;
static struct hrtimer timer;

static int init(void)
{
	int res;

	strcpy(v4l2_dev.name, "vdummy");
	if((res = v4l2_device_register(0, &v4l2_dev)) != 0) {
		return res;
	}

	mutex_init(&mutex);
	spin_lock_init(&blist_lock);

	vbops.queue_setup = queue_setup;
	vbops.buf_prepare = buf_prepare;
	vbops.buf_queue = buf_queue;
	vbops.start_streaming = start_streaming;
	vbops.stop_streaming = stop_streaming;
	vbops.wait_prepare = vb2_ops_wait_prepare;
	vbops.wait_finish = vb2_ops_wait_finish;

	vbq.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	vbq.io_modes = VB2_MMAP | VB2_READ;
	vbq.buf_struct_size = sizeof *buflist;
	vbq.ops = &vbops;
	vbq.mem_ops = &vb2_vmalloc_memops;
	vbq.timestamp_flags = V4L2_BUF_FLAG_TIMESTAMP_MONOTONIC;
	vbq.min_buffers_needed = 2;
	vbq.lock = &mutex;
	vbq.gfp_flags = GFP_KERNEL;

	if((res = vb2_queue_init(&vbq)) != 0) {
		printk(KERN_ALERT "vdummy: failed to initialize videobuf2 queue\n");
		v4l2_device_unregister(&v4l2_dev);
		return res;
	}

	if(!(vdev = video_device_alloc())) {
		return -ENOMEM;
	}
	memset(vdev, 0, sizeof *vdev);
	vdev->release = video_device_release;
	strcpy(vdev->name, KBUILD_MODNAME);
	vdev->fops = &fops;
	vdev->ioctl_ops = &iops;
	vdev->vfl_type = VFL_TYPE_GRABBER;
	vdev->vfl_dir = VFL_DIR_RX;
	vdev->v4l2_dev = &v4l2_dev;
	vdev->lock = &mutex;
	vdev->queue = &vbq;

	fops.owner = THIS_MODULE;
	fops.open = v4l2_fh_open;
	fops.release = vb2_fop_release;
	fops.read = vb2_fop_read;
	fops.mmap = vb2_fop_mmap;
	fops.poll = vb2_fop_poll;
	fops.unlocked_ioctl = video_ioctl2;

	iops.vidioc_querycap = ioctl_querycap;
	iops.vidioc_enum_fmt_vid_cap = ioctl_enum_fmt;
	iops.vidioc_enum_framesizes = ioctl_enum_framesz;
	iops.vidioc_g_fmt_vid_cap = ioctl_get_fmt;
	iops.vidioc_s_fmt_vid_cap = ioctl_set_fmt;
	iops.vidioc_try_fmt_vid_cap = ioctl_try_fmt;
	iops.vidioc_g_parm = ioctl_get_parm;
	iops.vidioc_s_parm = ioctl_set_parm;
	iops.vidioc_queryctrl = ioctl_queryctl;
	iops.vidioc_reqbufs = vb2_ioctl_reqbufs;
	iops.vidioc_create_bufs = vb2_ioctl_create_bufs;
	iops.vidioc_querybuf = vb2_ioctl_querybuf;
	iops.vidioc_qbuf = vb2_ioctl_qbuf;
	iops.vidioc_dqbuf = vb2_ioctl_dqbuf;
	iops.vidioc_expbuf = vb2_ioctl_expbuf;
	iops.vidioc_streamon = vb2_ioctl_streamon;
	iops.vidioc_streamoff = vb2_ioctl_streamoff;
	iops.vidioc_log_status = v4l2_ctrl_log_status;
	iops.vidioc_subscribe_event = v4l2_ctrl_subscribe_event;
	iops.vidioc_unsubscribe_event = v4l2_event_unsubscribe;

	if((res = video_register_device(vdev, VFL_TYPE_GRABBER, -1)) != 0) {
		printk(KERN_ALERT "vdummy: failed to register video device\n");
		v4l2_device_unregister(&v4l2_dev);
		video_device_release(vdev);
		return res;
	}

	hrtimer_init(&timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
	timer.function = timer_func;

	if((res = update_frame(640, 480)) != 0) {
		video_unregister_device(vdev);
		v4l2_device_unregister(&v4l2_dev);
		video_device_release(vdev);
		return res;
	}

	printk(KERN_INFO "vdummy: initialized\n");
	return 0;
}

static void shutdown(void)
{
	hrtimer_cancel(&timer);
	video_unregister_device(vdev);
	v4l2_device_unregister(&v4l2_dev);
	kfree(frame);
}

static int ioctl_querycap(struct file *file, void *fh, struct v4l2_capability *cap)
{
	strcpy(cap->driver, KBUILD_MODNAME);
	strcpy(cap->card, "Dummy V4L2 device");
	strcpy(cap->bus_info, "nobus");
	cap->device_caps = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE | V4L2_CAP_STREAMING;
	cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;
	return 0;
}

static int ioctl_enum_fmt(struct file *file, void *fh, struct v4l2_fmtdesc *fdesc)
{
	if(fdesc->index != 0 || fdesc->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}
	fdesc->flags = 0;
	strcpy(fdesc->description, "RGB24");
	fdesc->pixelformat = V4L2_PIX_FMT_RGB24;
	return 0;
}

static int ioctl_enum_framesz(struct file *file, void *fh, struct v4l2_frmsizeenum *fsz)
{
	if(fsz->index != 0) {
		return -EINVAL;
	}
	fsz->type = V4L2_FRMSIZE_TYPE_CONTINUOUS;
	fsz->stepwise.min_width = fsz->stepwise.min_height = 1;
	fsz->stepwise.max_width = fsz->stepwise.max_height = 8192;
	fsz->stepwise.step_width = fsz->stepwise.step_height = 1;
	return 0;
}

static int ioctl_get_fmt(struct file *file, void *fh, struct v4l2_format *fmt)
{
	if(fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}
	fmt->fmt.pix.width = width;
	fmt->fmt.pix.height = height;
	fmt->fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
	fmt->fmt.pix.field = V4L2_FIELD_NONE;
	fmt->fmt.pix.bytesperline = 0;
	fmt->fmt.pix.sizeimage = frame_size;
	fmt->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
	fmt->fmt.pix.priv = 0;
	fmt->fmt.pix.flags = 0;
	fmt->fmt.pix.ycbcr_enc = 0;
	fmt->fmt.pix.quantization = V4L2_QUANTIZATION_FULL_RANGE;
	fmt->fmt.pix.xfer_func = V4L2_XFER_FUNC_NONE;
	return 0;
}

static int ioctl_set_fmt(struct file *file, void *fh, struct v4l2_format *fmt)
{
	int res;

	if(fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}
	if((res = update_frame(fmt->fmt.pix.width, fmt->fmt.pix.height)) != 0) {
		return res;
	}
	return ioctl_get_fmt(file, fh, fmt);
}

static int ioctl_try_fmt(struct file *file, void *fh, struct v4l2_format *fmt)
{
	if(fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}
	if(fmt->fmt.pix.width <= 0)
		fmt->fmt.pix.width = width;
	if(fmt->fmt.pix.height <= 0)
		fmt->fmt.pix.height = height;
	fmt->fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
	fmt->fmt.pix.field = V4L2_FIELD_NONE;
	fmt->fmt.pix.bytesperline = 0;
	fmt->fmt.pix.sizeimage = frame_size;
	fmt->fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
	fmt->fmt.pix.priv = 0;
	fmt->fmt.pix.flags = 0;
	fmt->fmt.pix.ycbcr_enc = 0;
	fmt->fmt.pix.quantization = V4L2_QUANTIZATION_FULL_RANGE;
	fmt->fmt.pix.xfer_func = V4L2_XFER_FUNC_NONE;
	return 0;
}

static int ioctl_get_parm(struct file *file, void *fh, struct v4l2_streamparm *parm)
{
	if(parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}
	parm->parm.capture.capability = 0;
	parm->parm.capture.capturemode = 0;
	parm->parm.capture.timeperframe.numerator = 1;
	parm->parm.capture.timeperframe.denominator = framerate;
	parm->parm.capture.extendedmode = 0;
	parm->parm.capture.readbuffers = 1;
	return 0;
}

static int ioctl_set_parm(struct file *file, void *fh, struct v4l2_streamparm *parm)
{
	int new_framerate;
	if(parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
		return -EINVAL;
	}

	new_framerate = parm->parm.capture.timeperframe.denominator / parm->parm.capture.timeperframe.numerator;
	framerate = new_framerate < 1 ? 1 : (new_framerate > 60 ? 60 : new_framerate);

	return ioctl_get_parm(file, fh, parm);
}

static int ioctl_queryctl(struct file *file, void *fh, struct v4l2_queryctrl *ctl)
{
	return -EINVAL;
}

static int queue_setup(struct vb2_queue *vbq, unsigned int *nbuf,
		unsigned int *nplanes, unsigned int *sizes, ALLOC_CTX_TYPE **alloc_ctx)
{
	if(vbq->num_buffers + *nbuf < 2) {
		*nbuf = 2 - vbq->num_buffers;
	}

	*nplanes = 1;
	sizes[0] = frame_size;
	alloc_ctx[0] = 0;
	return 0;
}

static int buf_prepare(struct vb2_buffer *vb)
{
	if(vb2_plane_size(vb, 0) < frame_size) {
		printk(KERN_ALERT "vdummy: buffer too small\n");
		return -EINVAL;
	}
	return 0;
}

static void buf_queue(struct vb2_buffer *vb)
{
	struct buffer *buf = (struct buffer*)((char*)vb - offsetof(struct buffer, vb));

	spin_lock(&blist_lock);
	printk(KERN_INFO "vdummy: enqueue buffer\n");
	if(buflist) {
		buflist_tail->next = buf;
		buflist_tail = buf;
	} else {
		buflist = buflist_tail = buf;
	}
	buf->next = 0;
	spin_unlock(&blist_lock);
}

static void clear_queue(enum vb2_buffer_state st)
{
	struct buffer *list;

	printk(KERN_INFO "vdummy: clear_queue\n");

	spin_lock(&blist_lock);
	list = buflist;
	buflist = buflist_tail = 0;
	spin_unlock(&blist_lock);

	while(list) {
		struct buffer *buf = list;
		list = list->next;
		vb2_buffer_done(&buf->vb.vb2_buf, st);
	}
}

static int start_streaming(struct vb2_queue *vbq, unsigned int count)
{
	ktime_t frame_interval;

	streaming = 1;
	seqno = 0;

	/* calculate frame interval in nanoseconds */
	if(framerate <= 1) {
		frame_interval = ktime_set(1, 0);
	} else {
		frame_interval = ktime_set(0, 1000000000ul / (unsigned long)framerate);
	}

	printk(KERN_INFO "vdummy: start streaming (interval: %ld ms)\n",
			(long)ktime_to_ms(frame_interval));
	hrtimer_start(&timer, frame_interval, HRTIMER_MODE_REL);
	return 0;
}

static void stop_streaming(struct vb2_queue *vbq)
{
	printk(KERN_INFO "vdummy: stop streaming\n");
	hrtimer_cancel(&timer);

	streaming = 0;
	clear_queue(VB2_BUF_STATE_ERROR);
}

static int update_frame(int xsz, int ysz)
{
	unsigned char *tmp;
	int i, j;
	int new_fsz = xsz * ysz * 3;

	if(xsz == width && ysz == height) {
		return 0;
	}

	printk(KERN_INFO "vdummy: update_frame(%d, %d)\n", xsz, ysz);

	if(!frame || frame_size < new_fsz) {
		if(!(tmp = kmalloc(new_fsz, GFP_KERNEL))) {
			return -ENOMEM;
		}
		kfree(frame);
		frame_size = new_fsz;
		width = xsz;
		height = ysz;
		frame = tmp;
	}

	tmp = frame;
	for(i=0; i<height; i++) {
		for(j=0; j<width; j++) {
			int val = i ^ j;
			*tmp++ = val;
			*tmp++ = (val << 1) & 0xff;
			*tmp++ = (val << 2) & 0xff;
		}
	}

	return 0;
}

static int gen_frame(void)
{
	struct buffer *buf = 0;
	unsigned long flags;

	printk(KERN_INFO "vdummy: gen_frame\n");

	spin_lock_irqsave(&blist_lock, flags);
	if(buflist) {
		buf = buflist;
		buflist = buflist->next;
		if(!buflist) buflist_tail = 0;
	}
	spin_unlock_irqrestore(&blist_lock, flags);

	if(buf) {
		memcpy(vb2_plane_vaddr(&buf->vb.vb2_buf, 0), frame, frame_size);

		buf->vb.vb2_buf.timestamp = ktime_get_ns();
		printk(KERN_INFO "vdummy: copying frame: %d ts: %lu\n", seqno,
				(unsigned long)buf->vb.vb2_buf.timestamp);
		buf->vb.sequence = seqno++;
		buf->vb.field = V4L2_FIELD_NONE;
		vb2_buffer_done(&buf->vb.vb2_buf, VB2_BUF_STATE_DONE);
		vb2_set_plane_payload(&buf->vb.vb2_buf, 0, frame_size);
	}
	return 0;
}

enum hrtimer_restart timer_func(struct hrtimer *timer)
{
	printk(KERN_INFO "vdummy: timer!\n");
	//gen_frame();
	return streaming ? HRTIMER_RESTART : HRTIMER_NORESTART;
}
